"""
Initialization of the functions and modules cubic equation
"""

from .cubicequation import *
from .version       import __version__,__author__,__copyright__,__credits__,__license__,__maintainer__,__email__,__status__
