# pylint: disable=missing-module-docstring

# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
# https://stackoverflow.com/questions/1523427/what-is-the-common-header-format-of-python-files
# https://epydoc.sourceforge.net/manual-fields.html#module-metadata-variables

__version__     = "0.0.2"

__author__      = "Gien van den Enden"
__copyright__   = "Copyright 2024, Gien van den Enden"
__credits__     = [ ]
__license__     = "GPL"

__maintainer__  = "Gien van den Enden"
__email__       = "swvandenenden@gmail.com"
__status__      = "Development"
